#include <array>
#include <iostream>
#include <vector>

//
#include "glad/glad.h"
//
#include "GLFW/glfw3.h"
#include <cmath>

using namespace std;

struct Vec2f {
	//public members by default
	Vec2f() = default;
	Vec2f(float x, float y) : x(x), y(y) {}

	float x = 0.f;
	float y = 0.f;
};

Vec2f add(Vec2f a, Vec2f b) { return Vec2f(a.x + b.x, a.y + b.y); }
float dot(Vec2f a, Vec2f b) { return a.x*b.x + a.y*b.y; }
float length(Vec2f a) { return sqrt(dot(a, a)); };
Vec2f operator+(Vec2f a, Vec2f b) { return Vec2f(a.x + b.x, a.y + b.y); }
Vec2f operator-(Vec2f a, Vec2f b) { return Vec2f(a.x - b.x, a.y - b.y); }
Vec2f operator*(float s, Vec2f a) { return Vec2f(s*a.x, s*a.y); }
Vec2f operator*(Vec2f a, float s) { return Vec2f(s*a.x, s*a.y); }

std::ostream &operator<<(std::ostream &out, Vec2f const &a) {
	return out << ' ' << a.x << ' ' << a.y;
}

class Triangle {
	//Private members by default
	Vec2f m_verts[3];

public:
	Triangle() = default;
	Triangle(Vec2f a, Vec2f b, Vec2f c) {
		m_verts[0] = a;
		m_verts[1] = b;
		m_verts[2] = c;
	}

	//functions to get/read value
	Vec2f a() const { return m_verts[0]; }
	Vec2f b() const { return m_verts[1]; }
	Vec2f c() const { return m_verts[2]; }

	Vec2f operator[](int index) const { return m_verts[index]; }

	//functions to set value
	Vec2f &a() { return m_verts[0]; }
	Vec2f &b() { return m_verts[1]; }
	Vec2f &c() { return m_verts[2]; }

	Vec2f &operator[](int index) { return m_verts[index]; }
};

Vec2f midPoint(Vec2f a, Vec2f b) {
	return 0.5*(a + b);	// This one also works for midpoint (Can you show why?)
}

std::vector<Vec2f> nestedSquare(std::vector<Vec2f> sq) {
	std::vector<Vec2f> out;
	out.push_back(midPoint(sq[0], sq[1]));
	out.push_back(midPoint(sq[1], sq[2]));
	out.push_back(midPoint(sq[2], sq[3]));
	out.push_back(midPoint(sq[3], sq[0]));

	return out;
}

std::vector<Vec2f> outNestedSquares(int max_level) {
	std::vector<Vec2f> out;
	std::vector<Vec2f> sq;

	//initial Square points
	sq.push_back(Vec2f(-1, 1));
	sq.push_back(Vec2f(1, 1));
	sq.push_back(Vec2f(1, -1));
	sq.push_back(Vec2f(-1, -1));

	for (int j = 0; j < max_level; j++) {
		std::cout << "=== Level: " << j << " ===" << std::endl;

		//Square
		for (int i = 0; i < sq.size() - 1; i++) {
			out.push_back(sq[i]);
			out.push_back(sq[i + 1]);
		}
		out.push_back(sq[sq.size()-1]);
		out.push_back(sq[0]);

		sq = nestedSquare(sq);

		//Diamond
		for (int i = 0; i < sq.size() - 1; i++) {
			out.push_back(sq[i]);
			out.push_back(sq[i + 1]);
		}
		out.push_back(sq[sq.size() - 1]);
		out.push_back(sq[0]);

		sq = nestedSquare(sq);
	}

	return out;
}

std::vector<Triangle> sierpenskiTriangle(Triangle tri) {
	std::vector<Triangle> out;

	//Create vertices
	Vec2f a = tri.a();
	Vec2f b = tri.b();
	Vec2f c = tri.c();
	Vec2f d = midPoint(a, b);
	Vec2f e = midPoint(b, c);
	Vec2f f = midPoint(c, a);

	out.push_back(Triangle(a, d, f));
	out.push_back(Triangle(d, b, e));
	out.push_back(Triangle(f, e, c));

	return out;
}

std::vector<Triangle> outSierpenskiTriangle(int max_level) {
	std::vector<Triangle> tri;
	std::vector<Triangle> next;

	//initial Triangle points
	Vec2f v0(-1, -1);
	Vec2f v1(1, -1);

	//Find the top point in equilibrium triangle
	Vec2f d(cos((1.f / 3.f)*M_PI),	
		sin((1.f / 3.f)*M_PI));		//60 degrees per angle
	float t = length(v1 - v0);		//same length per side
	Vec2f v2 = v0 + t * d;

	//Base triangle
	tri.push_back(Triangle(v0, v1, v2));

	for (int j = 0; j < max_level; j++) {
		std::cout << "=== Level: " << j << " ===" << std::endl;

		//Sierpenski Triangle
		for (int i = 0; i < tri.size(); i++) {
			std::vector<Triangle> tmp = sierpenskiTriangle(tri[i]);
			next.insert(next.end(), tmp.begin(), tmp.end());
		}

		//Set up for next iteration
		tri.swap(next);
		next.clear();
	}

	return tri;
}


int main() {
  GLFWwindow *window = nullptr;

  // Close if OpenGL is not initialized
  if (!glfwInit()) {
    exit(EXIT_FAILURE);
  }
  
  // Currently installed version of OpenGL
  int glfwMajor, glfwMinor, glfwRevision;
  glfwGetVersion(&glfwMajor, &glfwMinor, &glfwRevision);
  std::cout << "GLFW Version: " << glfwMajor << "." << glfwMinor << "."
       << glfwRevision << std::endl;

  //Assign OpenGL version for program
  glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 2);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);

  //Create window
  window = glfwCreateWindow(800, 800, "Tut03", NULL, NULL);
  if (!window) {
    glfwTerminate();
    exit(EXIT_FAILURE);
  }

  //Set context to window
  glfwMakeContextCurrent(window);
  
  //Load up 
  gladLoadGLLoader((GLADloadproc)glfwGetProcAddress);

  glfwSwapInterval(1); // vsync
  
  //Set data
  
  // NDC [-1, 1]
  //std::vector<Vec2f> lines = outNestedSquares(5);
  std::vector<Triangle> tris = outSierpenskiTriangle(7);
 
  // Rendering loop
  while (!glfwWindowShouldClose(window)) {

    glClear(GL_COLOR_BUFFER_BIT);

	//Sierpinski Triangle
    glBegin(GL_TRIANGLES);
	glColor3f(1.f, 0.f, 1.f);
    for(int i = 0; i < tris.size(); i++) {
		Triangle t = tris[i];		
		glVertex2f(t[0].x, t[0].y);
		glVertex2f(t[1].x, t[1].y);
		glVertex2f(t[2].x, t[2].y);
	}    
    glEnd();
	
	/*
	//Nested Squares and Diamonds
	glBegin(GL_LINES);
	glColor3f(0.f, 1.f, 1.f);
	for (int i = 0; i < lines.size(); i++) {
		Vec2f v = lines[i];		
		glVertex2f(v.x, v.y);
	}
	glEnd();
	*/

	//Swap current scene with next scene
    glfwSwapBuffers(window);

    glfwPollEvents(); // will process event queue and carry on
  }

  glfwDestroyWindow(window);
  glfwTerminate();

  return EXIT_SUCCESS;
}
